/*
 * $RCSfile: MachineId.cc,v $
 * $Revision: 1.9 $
 *
 * Comments:
 *
 * (C) Copyright ParaSoft Corporation 1996.  All rights reserved.
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft
 * The copyright notice above does not evidence any
 * actual or intended publication of such source code.
 *
 * $Author: mali $          $Locker:  $
 * $Date: 2007-11-30 09:10:14 $
 * $Log: not supported by cvs2svn $
 * Revision 1.8  2007/11/29 14:59:49  dario
 * added tryHarder method
 *
 * Revision 1.7  2007/04/19 13:02:21  dario
 * eth2 linux interface added
 *
 * Revision 1.6  2006/11/28 12:37:07  dario
 * setErrorMessage implementation added
 *
 * Revision 1.5  2006/11/28 11:33:16  dario
 * getErrorMessage added
 *
 * Revision 1.4  2005/07/12 08:58:11  dario
 * unused functions removed
 *
 * Revision 1.3  2005/07/08 09:37:01  dario
 * macMachineId added
 *
 * Revision 1.2  2005/07/08 09:20:45  dario
 * parasoft header added.
 *
 */
 
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <linux/if.h>
#include <netinet/in.h> 
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ifaddrs.h>
#include <netdb.h>

#include "MachineId.h"


static const char *INTERFACES[] = { "eth", "tr", "wlan", "ppp", "bond", "vmnet", "em", "p1p", "p2p", "p3p", "p4p", "virbr", "ens3", 0};
static const char* POSTFIXES[] = { "", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", 0 };
static const char* ENV_NAME = "PARASOFT_SUPPORT_NET_INTERFACES";
static const char* ENV_DEBUG = ::getenv("PARASOFT_DEBUG_NET_INTERFACES");
static const char* ENV_DEBUG_FILE = ::getenv("PARASOFT_DEBUG_NET_INTERFACES_FILE");
static const bool IS_DEBUG_FILE = ENV_DEBUG_FILE != NULL && strcmp(ENV_DEBUG_FILE, "") != 0;
static const bool DEBUG = IS_DEBUG_FILE || (ENV_DEBUG != NULL && strcmp(ENV_DEBUG, "true") == 0);

static const char** getLinuxInterfaces();
static char* getDefaultInterface();
static char* getDefaultInterface(ifaddrs* ifaddr, int family);
static int calcMachineId(ifreq ifr);
static void printDebug(const char* format, ...);

void MachineId::setErrorMessage(const char* message)
{
    if (_errorMessage) {
        delete[] _errorMessage;
    }
    _errorMessage = new char[strlen(message) + 1];
    strcpy(_errorMessage, message);
}

int MachineId::getMachineId()
{
    const char **p = 0;
    int sd;
    struct ifreq ifr;
    sd = socket (PF_INET, SOCK_STREAM, 6);
    if (sd < 0) {
        setErrorMessage("Socket result less than 0");
        return 0x0;
    }
    memset(&ifr, 0, sizeof(struct ifreq));

    printDebug("====================================\n");
    const char** allInterfaces = getLinuxInterfaces();
    printDebug("Probing interfaces\n");
    for (p = allInterfaces; *p; p++) {
    	printDebug("Probing type %s\n", p[0]);
        for (const char** j = POSTFIXES; *j; j++) {
            if ((snprintf(ifr.ifr_name, sizeof(ifr.ifr_name), "%s%s", p[0], j[0]) < (int)sizeof(ifr.ifr_name)) &&
                (ioctl(sd, SIOCGIFHWADDR, &ifr) == 0) && 
                (calcMachineId(ifr) != 0) ) 
            {
		    	printDebug("[SELECTED] Accepted interface %s\n", ifr.ifr_name);
                goto SCANNING_FINISHED; //finish scanning, ifr is set
            }
        }
    }
 SCANNING_FINISHED:    
    
    close(sd);

    if (*p) {
        // at this point, ifr contains MAC address
        int machineId = calcMachineId(ifr);
        return machineId;
    }

    printDebug("None of the interfaces accepted, using backup procedure...\n");
    // try harder - following alex/license.cc
    int iTryHarderID = tryHarder(allInterfaces);
    if (iTryHarderID != 0) {
        return iTryHarderID;
    }

	printDebug("None of the interfaces matched! Will return 0\n");
    setErrorMessage("No linux interfaces");
    return 0x0;
}

// copy of tryHarder method from alex/license.cc for linux arch
int MachineId::tryHarder(const char* allInterfaces[])
{
    struct ifreq s;
    int fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
    const char **p = 0;
    
    if (fd < 0) {
        setErrorMessage("Socket result less than 0");
        return 0x0;
    }

    for (p = allInterfaces; *p; p++) {
        for (const char** j = POSTFIXES; *j; j++) {
            if (snprintf(s.ifr_name, sizeof(s.ifr_name), "%s%s", p[0], j[0]) < 
                (int)sizeof(s.ifr_name) &&
                0 == ioctl(fd, SIOCGIFADDR, &s)) {
                goto DONE;
            }
            // also try interface aliases, e.g. "eth0:1" (PR46394)
            for (int i = 1; i < 100; ++i) {
                if (snprintf(s.ifr_name, sizeof(s.ifr_name), "%s%s:%d", p[0], j[0], i) < 
                    (int)sizeof(s.ifr_name) &&
                    0 == ioctl(fd, SIOCGIFADDR, &s)) {
                    goto DONE;
                }
            }
        }
    }
 DONE:
    close(fd);
    if (*p) {
        // at this point, s contains local address in network byte format
        long addr = *(long *)&s.ifr_addr.sa_data[2];
        return (int)addr;
    }
    return 0;
}

static const char** getLinuxInterfaces()
{
    char* defaultVal = getDefaultInterface();
    char* envInterfaces = ::getenv(ENV_NAME);
    
    if ((defaultVal == 0) && (envInterfaces == 0)) {
    	printDebug("No default chosen, no envvar, using predefined list of known interfaces.\n");
        return INTERFACES;
    }

    char** result = new char*[30];
    int n = 0;

    // Add interfaces from environment variable
    if (envInterfaces != 0) {
        char* token;
        char* tofree = strdup(envInterfaces);
        while ((token = strsep(&tofree, ",")) != NULL) {
            result[n++] = strdup(token);
    		printDebug("Adding envvar interface %s\n", result[n-1]);
        }
        free(tofree);
    }

    // Add predefined interfaces first
    for (const char** p = INTERFACES; *p; p++) {
        result[n++] = strdup(p[0]);
		printDebug("Adding known interface %s\n", result[n-1]);
    }
    
    // Add default interface
    if (defaultVal != 0) {
        result[n++] = defaultVal;
		printDebug("Adding default interface %s\n", result[n-1]);
    }

    result[n] = 0;
    return (const char**)result;        
}

static char* getDefaultInterface()
{
    struct ifaddrs *ifaddr;
    if (getifaddrs(&ifaddr) == -1) {
        return NULL;
    }

    char* result = getDefaultInterface(ifaddr, AF_INET);
    if (result == NULL) {
        result = getDefaultInterface(ifaddr, AF_INET6);
    }
    freeifaddrs(ifaddr);
    return result;
}

static char* getDefaultInterface(ifaddrs* ifaddr, int family)
{
    int n;
    char* result = NULL;
    struct ifaddrs *ifa;
    
    printDebug("Determining default interface in family %d\n", family);

    for (ifa = ifaddr, n = 0; ifa != NULL; ifa =  ifa->ifa_next, n++) {
	    printDebug("Checking candidate %d: %s ...   \n", n+1, ifa->ifa_name);
        if (ifa->ifa_addr == NULL || ifa->ifa_name == NULL) {
            continue;
        }
        if ((ifa->ifa_flags & IFF_LOOPBACK) == IFF_LOOPBACK) {
	    	printDebug("   loopback - ignored\n");
            continue;
        }
        if (ifa->ifa_addr->sa_family != family) {
	    	printDebug("   family %d not matching - ignored\n", ifa->ifa_addr->sa_family);
            continue;
        }
        if ((result == NULL) || (strcmp(ifa->ifa_name, result) < 0)) {
            result = ifa->ifa_name;
		    printDebug("   accepted candidate %s\n", result);
        }
    }
    if (result == NULL) {
	    printDebug("None matched criteria for default interface\n");
        return NULL;
    }
    printDebug("Accepted as default: %s\n", result);
    return strdup(result);
}

static int calcMachineId(ifreq ifr)
{
    union U {
        unsigned char c[4];
        unsigned int  n;
    } u;

    unsigned char mac[6];
    memcpy(mac, &ifr.ifr_ifru.ifru_hwaddr.sa_data, 6);

 /*
    printf("MAC Address(%s): %02x:%02x:%02x:%02x:%02x:%02x\n", 
        ifr.ifr_name,
        (unsigned char)mac[0],
        (unsigned char)mac[1],
        (unsigned char)mac[2],
        (unsigned char)mac[3],
        (unsigned char)mac[4],
        (unsigned char)mac[5]);
*/

    unsigned int sum = 0;
    for (int ii = 0; ii < 6; ii++) {
        sum += mac[ii];
    }
    u.c[0] = mac[1];
    u.c[1] = mac[3];
    u.c[2] = mac[5]; 
    u.c[3] = (unsigned char)sum;
    return u.n;    
}

static void printDebug(const char* format, ...)
{
	if (!DEBUG) {
		return;
	}
	
    va_list argptr;
    va_start(argptr, format);
    
	if (IS_DEBUG_FILE) {
		FILE * debugFile = fopen(ENV_DEBUG_FILE, "a");
		if (debugFile != NULL) {
		    vfprintf(debugFile, format, argptr);
			fclose(debugFile);
		} else {
		    printf("[Wrong PARASOFT_DEBUG_NET_INTERFACES_FILE] %s\n", ENV_DEBUG_FILE);
		    vfprintf(stdout, format, argptr);
		}
	} else {
	    vfprintf(stdout, format, argptr);
	}
	
    va_end(argptr);
    
}
